@@ -1,14 +1,17 @@
 """Create the Awesome Panel Sharing App"""
 from __future__ import annotations
 
+from pathlib import Path
+
 import panel as pn
 
 from panel_sharing import components
-from panel_sharing.models import AppState
+from panel_sharing.models import AppState, AzureBlobStorage, Project, Site
 from panel_sharing.utils import (
     exception_handler,
     get_app_key,
     get_example_key,
+    get_project_key,
     notify_app_key_not_found,
 )
 
@@ -25,6 +28,45 @@ if not "panel-sharing" in pn.state.cache:
     pn.state.cache["panel-sharing"] = {"state": {}}
 
 
+@pn.cache
+def get_examples(examples: str):
+    """Returns a list of example Projects"""
+    # pylint: disable=protected-access
+    return list(components.Gallery.read(Path(examples))._examples_map.values())
+
+
+def _set_start_project(state, set_example, gallery):
+    key = get_app_key()
+    project = get_project_key()
+    example = get_example_key()
+    if key:
+        try:
+            state.set_dev_project_from_shared_app(key)
+        except:  # pylint: disable=bare-except
+            notify_app_key_not_found("app", key)
+            set_example(project=gallery.value)
+            pn.state.location.search = ""
+    elif project:
+        try:
+            project_ = Project.from_base64(project)
+            state.project.source.code = project_.source.code
+            state.project.source.requirements = project_.source.requirements
+            state.project.source.readme = project_.source.readme
+            state.project.source.thumbnail = project_.source.thumbnail
+            pn.state.onload(state.build)
+        except:  # pylint: disable=bare-except
+            notify_app_key_not_found("project", key)
+            set_example(project=gallery.value)
+            pn.state.location.search = ""
+    else:
+        try:
+            set_example(project=gallery.get(example))
+        except:  # pylint: disable=bare-except
+            notify_app_key_not_found("example", key)
+            set_example(project=gallery.value)
+            pn.state.location.search = ""
+
+
 def create():
     """Returns an instance of the Panel Sharing app
 
@@ -37,34 +79,19 @@ def create():
         "ace", sizing_mode="stretch_width", notifications=True, exception_handler=exception_handler
     )
 
-    state = AppState()
+    site = Site(production_storage=AzureBlobStorage())
+    state = AppState(site=site)
 
-    gallery = components.Gallery.read(state.examples)
+    gallery = components.Gallery(examples=get_examples(str(state.examples.absolute())))
 
     @pn.depends(gallery.param.value, watch=True)
     def set_example(project):
-        state.copy(project, source=state.examples / project.source.name)
+        project.build()
+        state.copy(project)
         pn.state.location.search = ""
         pn.state.location.update_query(example=project.name)
 
-    key = get_app_key()
-    example = get_example_key()
-    if key:
-        try:
-            state.set_project_from_app_key(key)
-            pn.state.location.search = ""
-            pn.state.location.update_query(app=key)
-        except:  # pylint: disable=bare-except
-            notify_app_key_not_found("app", key)
-            set_example(project=gallery.value)
-            pn.state.location.search = ""
-    else:
-        try:
-            set_example(project=gallery.get(example))
-        except:  # pylint: disable=bare-except
-            notify_app_key_not_found("example", key)
-            set_example(project=gallery.value)
-            pn.state.location.search = ""
+    _set_start_project(state, set_example, gallery)
 
     project_builder = components.ProjectBuilder(state=state)
     source_editor = components.SourceEditor(source=state.project.source)
